# Copyright (C) Patrick Brady, Brian Moe, Branson Stephens (2015)
#
# This file is part of lvalert
#
# lvalert is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# It is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lvalert.  If not, see <http://www.gnu.org/licenses/>.

#from ligo.lvalert.sleeklvalert import * 

from .version import __version__

import sys
import getpass
import logging
import netrc
import uuid
import ssl

import pkg_resources
from ligo.lvalert.utils import safe_netrc as _netrc

import sleekxmpp

# Experimental error trapping:
import functools

__all__ = ('LVAlertClient',)

log = logging.getLogger(__name__)

DEFAULT_SERVER = 'lvalert.cgca.uwm.edu'

# Python versions before 3.0 do not use UTF-8 encoding
# by default. To ensure that Unicode is handled properly
# throughout SleekXMPP, we will set the default encoding
# ourselves to UTF-8.
if sys.version_info < (3, 0):
    reload(sys)
    sys.setdefaultencoding('utf8')
else:
    raw_input = input

def _get_default_login(netrcfile, server):
    try:
        netrcfile = _netrc(netrcfile)
    except (OSError, netrc.NetrcParseError) as e:
        log.exception('Cannot load netrc file (%s): %s', netrcfile, e)
        return None, None

    auth = netrcfile.authenticators(server)
    if auth is None:
        log.warn('No netrc entry found for server: %s', server)
        return None, None

    default_username, _, default_password = auth
    return default_username, default_password


def _get_login(username, password, netrc, interactive, server):
    default_username, default_password = _get_default_login(netrc, server)
    prompt = 'password for {}@{}: '.format(username, server)

    if username is not None and password is not None:
        return username, password
    elif username is None and default_username is None:
        raise RuntimeError('Username not specified')
    elif username is None or username == default_username:
        return default_username, default_password
    elif interactive:
        return username, getpass.getpass(prompt)
    else:
        raise RuntimeError('Password not specified')

#-- The purpose of this function is to catch exceptions from 
#-- external functions before they get caught in the xml stream
def catch_exception(f):
    @functools.wraps(f)
    def func(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception as e:
            print('Caught an exception in ', f.__name__, ":")
            print(e)
    return func


class LVAlertClient(sleekxmpp.ClientXMPP):
    """An XMPP client configured for LVAlert.

    Parameters
    ----------
    username : str (optional)
        The XMPP username, or :obj:`None` to look up from the netrc_ file.
    password : str (optional)
        The XMPP password, or :obj:`None` to look up from the netrc_ file.
    resource : str (optional)
        The XMPP resource ID, or :obj:`None` to generate a random one.
    netrc : str (optional)
        The netrc_ file. The default is to consult the ``NETRC`` environment
        variable or use the default path of ``~/.netrc``.
    interactive : bool (optional)
        If :obj:`True`, then fall back to asking for the password on the
        command line if necessary.
    server : str (optional)
        The LVAlert server hostname.

    Example
    -------

    Here is an example for performing administrative actions.

    .. code-block:: python

        client = LVAlertClient()
        client.connect()
        client.process(block=False)
        client.subscribe('cbc_gstlal', 'cbc_pycbc')
        client.abort()

    Here is an example for running a listener.

    .. code-block:: python

        def process_alert(node, payload):
            if node == 'cbc_gstlal':
                alert = json.loads(payload)
                ...

        client = LVAlertClient()
        client.listen(process_alert)
        client.connect()
        client.process(block=True)
    """

    def __init__(self, username=None, password=None, resource=None, netrc=None,
                 interactive=False, server=DEFAULT_SERVER):
        username, password = _get_login(
            username, password, netrc, interactive, server)
        if resource is None:
            resource = uuid.uuid4().hex
        jid = '{}@{}/{}'.format(username, server, resource)

        super(LVAlertClient, self).__init__(jid, password)

	# Activate PubSub plugin
        self.register_plugin('xep_0060')

        # Test out xmpp ping to see if it does anything:
        self.register_plugin('xep_0199', {'keepalive': True, 'interval': 5, 'timeout': 5})

        # Add handlers for available events. 
        self.add_event_handler("session_start", self._session_start, threaded=True)

	# Point the clients to the certs. This needs to be addressed.
        self.ca_certs = pkg_resources.resource_filename(__name__, 'certs.pem')

    def _session_start(self, event):
        self.get_roster()
        self.send_presence()

    def listen(self, callback):
        """Set a callback to be executed for each pubsub item received.

        Parameters
        ----------
        callback : callable
            A function of two arguments: the node and the alert payload.
        """
        self._callback = callback
        self.add_event_handler('pubsub_publish', self._pubsub_publish)

    @catch_exception
    def _pubsub_publish(self, msg):
        if msg['type'] in ('chat', 'normal'):
            self._callback(msg['pubsub_event']['items']['node'],
                           msg['pubsub_event']['items']['item']['payload'].text)

    @property
    def _pubsub_server(self):
        return 'pubsub.{}'.format(self.boundjid.server)

    def get_nodes(self):
        """Get a list of all available pubsub nodes."""
        result = self['xep_0060'].get_nodes(self._pubsub_server)
        return [item for _, item, _ in result['disco_items']['items']]

    def get_subscriptions(self):
        """Get a list of your subscriptions."""
        result = self['xep_0060'].get_subscriptions(self._pubsub_server)
        return sorted({stanza['node'] for stanza in
                       result['pubsub']['subscriptions']['substanzas']})

    def subscribe(self, *nodes):
        """Subscribe to one or more pubsub nodes."""
        for node in nodes:
            log.info('Subscribing to %s', node)
            self['xep_0060'].subscribe(self._pubsub_server, node)

    def unsubscribe(self, *nodes):
        """Unsubscribe from one or more pubsub nodes."""
        for node in nodes:

            subscriptions = self['xep_0060'].get_subscriptions(
                self._pubsub_server, node
            )['pubsub']['subscriptions']['substanzas']

            for subscription in subscriptions:
                log.info('Unsubscribing from %s [%s]',
                         node, subscription['subid'])
                self['xep_0060'].unsubscribe(
                    self._pubsub_server, node, subscription['subid'])

    def publish(self, node, msg=None):
        """Publish a message to one or more pubsub nodes."""
        xmlmsg = sleekxmpp.ET.fromstring("<pubsub xmlns='http://jabber.org/protocol/pubsub'>%s</pubsub>"
                                             % msg)
        try:
            result = self['xep_0060'].publish(self._pubsub_server, node,
                                              payload=xmlmsg)
            id = result['pubsub']['publish']['item']['id']
            log.info('Published at item id: %s' % id)
        except Exception as e:
            print(e)
            log.error('Could not publish to: %s' % node)

    def delete(self):
        try:
            self['xep_0060'].delete_node(self.pubsub_server, self.node)
            print('Deleted node: %s' % self.node)
        except:
            logging.error('Could not delete node: %s' % self.node)



# Copyright (C) Patrick Brady, Brian Moe, Branson Stephens (2015)
#
# This file is part of lvalert
#
# lvalert is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# It is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lvalert.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function
import os
import time
import logging
import libxml2
import getpass
import uuid
from optparse import *

# pubsub import must come first because it overloads part of the
# StanzaProcessor class
from ligo.lvalert import pubsub
from ligo.lvalert.utils import safe_netrc

from pyxmpp.all import JID,Iq,Presence,Message,StreamError,TLSSettings
from pyxmpp.jabber.all import Client
from pyxmpp.jabber.dataforms import Form,DATAFORM_NS
from pyxmpp.interface import implements
from pyxmpp.interfaces import *
from pyxmpp.exceptions import ClientError, FatalClientError

"""
A tool to administer the pubsub service of openfire
"""

#################################################################
# help message
usage = """\
%prog [options]
------------------------------------------------------------------------------
  A tool to administer the pubsub service of openfire at
  lvalert.cgca.uwm.edu. LSC-Virgo members can activate their accounts 
  on this server by completing the form at
  
    https://www.lsc-group.phys.uwm.edu/cgi-bin/jabber-acct.cgi 

  and typing your password. Then you can use this program to do the
  following things:

  1.  create a pubsub node called small_steps

  %prog --username albert.einstein --create --node small_steps

  2.  delete a pubsub node called small_steps

  %prog --username albert.einstein --delete --node small_steps

  4.  subscribe to a pubsub node called lvalert_test

  %prog --username albert.einstein --subscribe --node lvalert_test

  The owner (person who creates the node by default) can delete and
  publish information to the node. The owner can also add other
  publishers to the node using the command
  
  %prog --username albert.einstein --add-publisher patrick.brady --node small_steps
  
  Others can subscribe. To see a list of available nodes:
  
  %prog --username albert.einstein --get-nodes 

  To see the nodes to which you are subscribed:

  %prog --username albert.einstein --subscriptions
  
  There are two programs called lvalert_listen and lvalert_send which
  provide ways to listen for content on the node and to publish
  information to the node respectively.

"""

#################################################################
parser = OptionParser(usage=usage)

#username and password
parser.add_option("-a","--username",action="store",type="string",\
  help="the username of the publisher or listener" )
parser.add_option("-s","--server",action="store",type="string",\
  default="lvalert.cgca.uwm.edu", help="the pubsub server" )
parser.add_option("-r","--resource",action="store",type="string",\
  default=uuid.uuid4().hex, help="resource to use in JID [default: random]" )
parser.add_option("--netrc", "-N",
  help="Load username and password from this file",
  default=os.environ.get('NETRC', '~/.netrc'))

# access information about root nodes
parser.add_option("-d","--create",action="store_true",\
  default=False, help="create a new pubsub node" )
parser.add_option("-e","--delete",action="store_true",\
  default=False, help="delete a new pubsub node" )
parser.add_option("-f","--subscribe",action="store_true",\
  default=False, help="subscribe to a new pubsub node" )
parser.add_option("-g","--unsubscribe",action="store_true",\
  default=False, help="unsubscribe from a new pubsub node" )
parser.add_option("-n","--unsubscribe-subid",action="store",type="string",\
  default=None, help="subid of subscription obtained by running subscriptions" )
parser.add_option("-i","--subscriptions",action="store_true",\
  default=False, help="return a list of subscriptions" )
parser.add_option("-j","--add-publisher",action="store",type="string",\
  default=None, help="jabber id of person allowed to publish to node" )
parser.add_option("-m","--get-nodes",action="store_true",\
  default=False, help="get the list of existing nodes" )
parser.add_option("-o","--affiliations",action="store",type="string",\
  default=None, help="get the list of affiliations to a node" )
parser.add_option("-p","--delete-publisher",action="store",type="string",\
  default=None, help="jid of publisher to delete from a node" )
parser.add_option("-q","--node",action="store",type="string",\
  default=None, help="name of node" )
# debugging options
parser.add_option("-l","--debug",action="store_true",\
  default=False, help="should  print out lots of information" )
parser.add_option("-v","--verbose",action="store_true",\
  default=False, help="be verbose as you process the request" )
# version
parser.add_option("-w", "--version", action="store_true",
  default=False, help="display version information")
  
# ============================================================================
# -- get command line arguments
opts, args = parser.parse_args()

if opts.version:
    import pkg_resources
    version = pkg_resources.require("ligo-lvalert")[0].version
    print("LVAlert v. %s" % version)
    exit(0)

if (opts.add_publisher or opts.create or opts.delete or opts.subscribe or opts.unsubscribe or opts.delete_publisher) and not opts.node:
    parser.error("--node must be provided")

try:
    default_username, _, default_password = safe_netrc(os.path.expanduser(opts.netrc)).authenticators(opts.server)
except IOError:
    default_username, default_password = None, None

myusername = opts.username or default_username
if not myusername:
    parser.error('--username is required')
if myusername == default_username:
    mypassword = default_password
else:
    mypassword = getpass.getpass('password for ' + myusername + ': ')

class MyClient(Client):
    def __init__(self, jid, password):
        # we require a TLS connection
        t=TLSSettings(require=True,verify_peer=False)

        # setup client with provided connection information
        # and identity data
        Client.__init__(self, jid, password, \
            auth_methods=["sasl:GSSAPI","sasl:PLAIN"], tls_settings=t)


    def stream_state_changed(self,state,arg):
        """This one is called when the state of stream connecting the
        component to a server changes. This will usually be used to
        let the user know what is going on."""
        if opts.verbose:
            print("*** State changed: %s %r ***" % (state,arg))
        else:
            pass

    def session_started(self):
        self.stream.send(Presence())
        if opts.get_nodes:
            self.stream.set_response_handlers(pspl, \
                pspl.get_nodes_result,pspl.create_error,\
                pspl.create_timeout)
        if opts.create:
            self.stream.set_response_handlers(pspl, \
                pspl.generic_result,pspl.create_error,\
                pspl.create_timeout)
        if opts.delete:
            self.stream.set_response_handlers(pspl, \
                pspl.generic_result,pspl.generic_error,\
                pspl.generic_timeout)
        if opts.subscribe:
            self.stream.set_response_handlers(pspl, \
                pspl.subscribe_result,pspl.subscribe_error,\
                pspl.subscribe_timeout)
        if opts.affiliations:
            self.stream.set_response_handlers(pspl, \
                pspl.affiliations_result,pspl.generic_error,\
                pspl.generic_timeout)
        if opts.add_publisher:
            self.stream.set_response_handlers(pspl, \
                pspl.generic_result,pspl.generic_error,\
                pspl.generic_timeout)
        if opts.delete_publisher:
            self.stream.set_response_handlers(pspl, \
                pspl.generic_result,pspl.generic_error,\
                pspl.generic_timeout)
        if opts.unsubscribe:
            self.stream.set_response_handlers(pspl, \
                pspl.generic_result,pspl.generic_error,\
                pspl.generic_timeout)
        if opts.subscriptions:
            self.stream.set_response_handlers(pspl, \
                pspl.subscriptions_result,pspl.subscribe_error,\
                pspl.subscribe_timeout)
        self.stream.send(pspl)

    def idle(self):
        if self.stream and self.session_established:
            if opts.verbose:
                print("Disconnecting")
            self.disconnect()
        if opts.verbose:
            print("idle")
        time.sleep(4)

    def post_disconnect(self):
        print("Disconnected")
        raise Disconnected


# add a logger so that we can see what's going
if opts.debug:
    logger=logging.getLogger()
    logger.addHandler(logging.StreamHandler())
    logger.setLevel(logging.DEBUG)
else:
    logging.basicConfig(level=logging.ERROR)

# debug the memore
libxml2.debugMemory(1)

# set up the stream
# append UUID to resource to guarantee that it is non-empty and unique
myjid=JID(myusername+"@"+opts.server+"/"+opts.resource)
s=MyClient(jid=myjid,password=mypassword)

if opts.verbose:
    print("connecting...")
s.connect()

if opts.verbose:
    print("build pubsub stanza...")
recpt=JID("pubsub."+opts.server)
pspl=pubsub.PubSub(from_jid = myjid, to_jid = recpt, stream = s,\
stanza_type="get")
if opts.get_nodes:
    print("Getting nodes")
    pspl.get_nodes()
elif opts.create:
    print("Creating node " + opts.node)
    pspl.create_node(opts.node)
elif opts.delete:
    pspl.delete_node(opts.node)
elif opts.subscribe:
    pspl.subscribe(myjid,opts.node)
elif opts.unsubscribe:
    pspl.unsubscribe(myjid,opts.node,opts.unsubscribe_subid)
elif opts.subscriptions:
    pspl.subscriptions(myjid)
elif opts.affiliations:
    pspl.affiliations(myjid,opts.affiliations)
elif opts.add_publisher:
    publisher_jid = opts.add_publisher.split('@')[0]
    publisher_jid=JID(publisher_jid+"@"+opts.server+"/Home")
    pspl.publisher(publisher_jid,opts.node,"publisher")
elif opts.delete_publisher:
    publisher_jid = opts.delete_publisher.split('@')[0]
    publisher_jid=JID(publisher_jid+"@"+opts.server+"/Home")
    pspl.publisher(publisher_jid,opts.node,"none")
else:
    pspl.get_nodes()

if opts.verbose:
    print("sending message...")

try:
    s.loop(1)
except KeyboardInterrupt:
    print(u"disconnecting...")
    s.disconnect()

# vi: sts=4 et sw=4



print("import: 'ligo.lvalert'")
import ligo.lvalert

print("import: 'ligo.lvalert.tool'")
import ligo.lvalert.tool

print("import: 'ligo.lvalert.utils'")
import ligo.lvalert.utils

print("import: 'ligo.lvalert.version'")
import ligo.lvalert.version




# Copyright (C) Patrick Brady, Brian Moe, Branson Stephens (2015)
# Copyright (C) Leo Singer (2016)
#
# This file is part of lvalert
#
# lvalert is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# It is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lvalert.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function
import sys
import os
import datetime
import logging
import libxml2
import getpass
import uuid
import six.moves.configparser
from subprocess import Popen, PIPE
from optparse import *

# pubsub import must come first because it overloads part of the
# StanzaProcessor class
from ligo.lvalert import pubsub
from ligo.lvalert.utils import safe_netrc

from pyxmpp.all import JID,Iq,Presence,Message,StreamError,TLSSettings
from pyxmpp.jabber.all import Client
from pyxmpp.jabber.simple import send_message
from pyxmpp.interface import implements
from pyxmpp.interfaces import *

"""
A tool to listen for events on a pubsub node 
"""

#################################################################
# help message
usage = """\
%prog [options]
-----------------------------------------------------------------

  A tool to listen to the pubsub service of openfire at
  lvalert.cgca.uwm.edu. LSC-Virgo members can activate their accounts 
  on this server by completing the form at
  
    https://www.lsc-group.phys.uwm.edu/cgi-bin/jabber-acct.cgi 

  and typing your password.

  Before using this program to listen to a node, you must subscribe to
  the node using lvalert_admin. Then you will receive any events that are
  published to that node by doing:

  %prog --username albert.einstein

  If you do not provide the username, then the program will look for it in
  the file ~/.netrc in your home directory. This location may be overriden
  with the NETRC environment variable or the --netrc option. This file
  should contain a line like the following:

      machine lvalert.cgca.uwm.edu login albert.einstein password ligorocks

  If a matching entry in your .netrc file is found, the program will use
  that password. Otherwise, it will prompt you interactively for your
  password.

  When an event is published to the node, a message will be printed to
  the window where the listener is running. To see the event, run

  %prog --username albert.einstein --show

  The owner (person who creates the node by default) can delete and
  publish information to the node. The owner can also add other
  publishers to the node. Configuration and management of nodes and
  subscriptions are handled with lvalert_admin. 
  
  Others can subscribe to any existing node. Run 

  lvalert_admin --help

  to find out how to manage your subscriptions. 

  It is also possible to specify actions to be taken upon receipt of a
  message from a given node using a config-file to specify the program
  to run when the message is received via different nodes. The payload
  of the message is piped to the command via stdin.  A sample
  config-file called example.ini might look like

  [lvalert_test]
  executible=./mycounter

  then running

  lvalert_listen --username albert.einstein secret --config-file example.ini

  would result in ./mycounter being executed and the output printed to
  screen whenever a message was received via the lvalert_test node. 

  Alternatively, instead of the name of an executable, you could
  indicate either "stdout" or "-" (withouth quotes) and the alert
  payload will be written to standard output.
"""

#################################################################
parser = OptionParser(usage=usage)

#username and password
parser.add_option("-a","--username",action="store",type="string",\
  help="the username of the publisher or listener" )
parser.add_option("-s","--server",action="store",type="string",\
  default="lvalert.cgca.uwm.edu", help="the pubsub server" )
parser.add_option("-r","--resource",action="store",type="string",\
  default=uuid.uuid4().hex, help="resource to use in JID [default: random]" )
parser.add_option("-c","--config-file",action="store",type="string",\
  default=None, help="config file with list of actions" )
parser.add_option("--netrc", "-N",
  help="Load username and password from this file",
  default=os.environ.get('NETRC', '~/.netrc'))

parser.add_option("-S","--show",action="store_true",\
  default=False, help="print the payload to stdout" )

parser.add_option("-n","--node",action="store",type="string",\
  default=None, help="name of the node on the pubsub server" )

# debugging options
parser.add_option("-v","--verbose",action="store_true",\
  default=False, help="be verbose as you process the request" )
parser.add_option("-g","--debug",action="store_true",\
  default=False, help="should  print out lots of information" )

# options about message handling
parser.add_option("", "--dont-wait",action="store_true",\
  default=False, help="if supplied, jobs will be launched \
  as soon as they arrive, rather than blocking")

# version
parser.add_option("-w", "--version", action="store_true",
  default=False, help="display version information")

# ============================================================================
# -- get command line arguments
opts, args = parser.parse_args()

if opts.version:
    import pkg_resources
    version = pkg_resources.require("ligo-lvalert")[0].version
    print("LVAlert v. %s" % version)
    exit(0)

try:
    default_username, _, default_password = safe_netrc(os.path.expanduser(opts.netrc)).authenticators(opts.server)
except IOError:
    default_username, default_password = None, None

myusername = opts.username or default_username
if not myusername:
    parser.error('--username is required')
if myusername == default_username:
    mypassword = default_password
else:
    mypassword = getpass.getpass('password for ' + myusername + ': ')

if opts.dont_wait: ### required packages for "dont-wait" option
    import multiprocessing
    import tempfile

#=============================================================================
### R. ESSICK 12/2/2014
def forked_wait(cmd, file_obj):
    """
    used with the "--dont-wait" option to avoid zombie processes via a double fork
    main process will wait for this function to finish (quick), send the "wait" signal
    which removes this function from the process table, and then moves on.
    the forked process that this function creates become orphaned, and will automatically
    be removed from the process table upon completion.
    """
    Popen(cmd, stdin=file_obj)

#=======================

class LVAlertHandler(object):
    """Provides the actions taken when an event arrives.
    """

    implements(IMessageHandlersProvider)
    
    def __init__(self, client, actions, wait=True):
        """Just remember who created this."""
        self.client = client
        self.actions = actions
        self.wait = wait
    
    def get_message_handlers(self):
        """Return list of (message_type, message_handler) tuples.

        The handlers returned will be called when matching message is received
        in a client session."""
        return [
            (None, self.message),
            ]

    def message(self,stanza):
        """Message handler for the component.

        Prints a message with the time an alert is received. If the
        --show option is set, then it will also print the contents of
        the alert.

        if self.wait: blocks until this processes terminates and prints the returned string
        else: does not block, and multiple events will be launched and processed as soon as they are read

        :returns: `True` to indicate, that the stanza should not be processed
        any further."""
        e=self.get_entry(stanza)
        n=self.get_node(stanza)
        if e:
            if n in self.actions:
                action = self.actions[n]
                if action == 'stdout' or action == '-':
                    sys.stdout.write(str(e))
                #===================================================================
                ### changes made by R. Essick (ressick@mit.edu) and R. Vaulin (vaulin@ligo.org)
                ###
                ### this avoids using p.communicate(), which blocks and prevents the 
                ### alert handler from launching another processes until this one finishes
                ### we use "spooled" tempfiles because they exist only in memory (if small enought)
                ### max_size=1000 (bytes?) should be larger than most expected lvalert messages?
                elif self.wait:
                    p = Popen([self.actions[n]], stdin=PIPE, stdout=PIPE)
                    print(p.communicate(e)[0])
                else:
#                    import tempfile
                    file_obj = tempfile.SpooledTemporaryFile(mode="w+r", max_size=1000)
                    file_obj.write(e)
                    file_obj.readlines() ### bug fix for "alert_type"=="new" events
                                         ### without this, the position in file_obj gets messed up
                                         ### no idea why, but it might be related to long messages becoming multiple stanzas
                    file_obj.seek(0, 0)
#                    p = Popen([self.actions[n]], stdin=file_obj)
                    p = multiprocessing.Process(target=forked_wait, args=([self.actions[n]], file_obj))
                    p.start()
                    p.join()
                    file_obj.close()
                #===================================================================
                sys.stdout.flush()
            else:
                print("Payload received at %s" % (datetime.datetime.now().ctime()))
                if opts.show:
                    print(u'%s' % (e,), end=' ')
        return True

    def get_node(self,stanza):
        c = stanza.xmlnode.children
        c = c.children
        if c:
            return c.prop("node")

    def get_entry(self,stanza):
        c = stanza.xmlnode.children
        while c:
            try:
                if c.name=="event":
                    return c.getContent()
            except libxml2.treeError:
                pass
            c = c.next
        return None

class MyClient(Client):
    def __init__(self, jid, password,actions, wait=True):
        # we require a TLS connection
        t=TLSSettings(require=True,verify_peer=False)

        # setup client with provided connection information
        # and identity data
        Client.__init__(self, jid, password, \
            auth_methods=["sasl:GSSAPI","sasl:PLAIN"], tls_settings=t,keepalive=30)

        # add the separate components
        self.interface_providers = [
            LVAlertHandler(self,actions, wait=wait),
            ]

    def stream_state_changed(self,state,arg):
        """This one is called when the state of stream connecting the
        component to a server changes. This will usually be used to
        let the user know what is going on."""
        if opts.verbose:
            print("*** State changed: %s %r ***" % (state,arg))
        else:
            pass

# add a logger so that we can see what's going
if opts.debug:
    logger=logging.getLogger()
    logger.addHandler(logging.StreamHandler())
    logger.setLevel(logging.DEBUG)
else:
    logging.basicConfig(level=logging.ERROR)

# debug the memory
libxml2.debugMemory(1)

# set up handlers for each node
actions={}
if opts.config_file:
    cp=six.moves.configparser.ConfigParser()
    cp.read(opts.config_file)
    for node in cp.sections():
        try:
            actions[node] = cp.get(node,'executible')
        except six.moves.configparser.NoOptionError:
            actions[node] = cp.get(node,'executable')

# set up the stream
# append UUID to resource to guarantee that it is non-empty and unique
myjid=JID(myusername+"@"+opts.server+"/"+opts.resource)
s=MyClient(myjid,mypassword,actions, wait=not opts.dont_wait)

if opts.verbose:
    print("connecting...")
s.connect()

if opts.verbose:
    print("listening for message...")
try:
    s.loop(1)
except KeyboardInterrupt:
    print(u"disconnecting...")
    s.disconnect()

# vi: sts=4 et sw=4



# Copyright (C) Patrick Brady, Brian Moe, Branson Stephens (2015)
#
# This file is part of lvalert
#
# lvalert is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# It is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lvalert.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function
import sys
import os
import datetime
import logging
import libxml2
import getpass
import uuid
from optparse import OptionParser

# pubsub import must come first because it overloads part of the
# StanzaProcessor class
from ligo.lvalert import pubsub
from ligo.lvalert.utils import safe_netrc

from pyxmpp.all import JID,TLSSettings
from pyxmpp.jabber.all import Client

"""
A tool to send an event to a pubsub node 
"""

__name__ = "lvalert_send"

#################################################################
# help message
usage = """\
%prog [options]
-----------------------------------------------------------------

  A tool to send to the pubsub service at lvalert.cgca.uwm.edu.
  LSC-Virgo members can activate their accounts on this server by
  completing the form at
  
    https://www.lsc-group.phys.uwm.edu/cgi-bin/jabber-acct.cgi 

  and setting a password.  Before using this program to send to a
  node, the node should be created with lvalert_admin and/or
  configured to allow you to send to it. 

  You can send information containted in the file stuff.xml by
  doing:

  %prog --username albert.einstein --node small_steps --file stuff.xml --max_attempts N

  When an stuff.xml is sent to a node by a publisher, it will be sent
  to all subscribers who are logged in.

  The owner (person who creates the node by default) can delete and
  publish information to the node. The owner can also add other
  publishers to the node. Configuration and management of nodes is
  handled with lvalert_admin. 

  If the server is busy, the connection is allowed to timeout N
  times before failing.
  
  Others can subscribe any existing node. Run 

  lvalert_admin --help

  to find out how to manage your subscriptions. 

"""

#################################################################
parser = OptionParser(usage=usage)

#username and password
parser.add_option("-a","--username",action="store",type="string",\
  help="the username of the publisher or listener" )
parser.add_option("-s","--server",action="store",type="string",\
  default="lvalert.cgca.uwm.edu", help="the pubsub server" )
parser.add_option("-r","--resource",action="store",type="string",\
  default=uuid.uuid4().hex, help="resource to use in JID [default: random]" )
parser.add_option("--netrc", "-N",
  help="Load username and password from this file",
  default=os.environ.get('NETRC', '~/.netrc'))

# access information about root nodes
parser.add_option("-n","--node",action="store",type="string",\
  default=None, help="name of the node on the pubsub server" )
parser.add_option("-p","--file",action="store",type="string",\
  default=None, help="name of the file with the event" )

# debugging options
parser.add_option("-v","--verbose",action="store_true",\
  default=False, help="be verbose as you process the request" )
parser.add_option("-g","--debug",action="store_true",\
  default=False, help="should  print out lots of information" )

# timeout attempts
parser.add_option("-m","--max_attempts",action="store_true",\
  default=10, help="max number of timeouts allowed" )

# version
parser.add_option("-w", "--version", action="store_true",
  default=False, help="display version information")

# ============================================================================
# -- get command line arguments
opts, args = parser.parse_args()

if opts.version:
    import pkg_resources
    version = pkg_resources.require("ligo-lvalert")[0].version
    print("LVAlert v. %s" % version)
    exit(0)

try:
    default_username, _, default_password = safe_netrc(os.path.expanduser(opts.netrc)).authenticators(opts.server)
except IOError:
    default_username, default_password = None, None

myusername = opts.username or default_username
if not myusername:
    parser.error('--username is required')
if myusername == default_username:
    mypassword = default_password
else:
    mypassword = getpass.getpass('password for ' + myusername + ': ')

if not opts.node:
    raise ValueError("--node is required")

if not opts.file:
    raise ValueError("--file is required")

# add a logger so that we can see what's going
logger=logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
if opts.debug:
    logger.setLevel(logging.DEBUG)
elif opts.verbose:
    logger.setLevel(logging.INFO)
else:
    logger.setLevel(logging.ERROR)

def onStart(client):
    node = opts.node

    if opts.file == '-':
        eventfile = sys.stdin
    else:
        eventfile = open(opts.file)
    voevent = eventfile.read()
    eventfile.close()

    recpt=JID("pubsub."+opts.server)

    client.sendMessage(node, voevent, recpt)

class MyClient(Client):
    def __init__(self,jid,password,max_attempts,onStart):
        self.jid = jid
        self.max_attempts = max_attempts
        self.onStart = onStart

        # we require a TLS connection
        t=TLSSettings(require=True,verify_peer=False)

        # setup client with provided connection information
        # and identity data
        Client.__init__(self, self.jid, password, \
            auth_methods=["sasl:GSSAPI","sasl:PLAIN"], tls_settings=t)

        # A counter for the idle function, below.
        self.counter = 0

    def stream_state_changed(self,state,arg):
        """This one is called when the state of stream connecting the component
        to a server changes. This will usually be used to let the user
        know what is going on."""
        logger.info("*** State changed: %s %r ***" % (state,arg))

    def session_started(self):
        logger.info(datetime.datetime.now().ctime())
        self.onStart(self)

    def sendMessage(self,node,msg,recpt):
        ps = pubsub.PubSub(from_jid = self.jid, to_jid = recpt, stream = self.stream,
                           stanza_type = "get")
        ps.publish(msg,node)
        self.stream.set_response_handlers(ps, 
                self.onSuccess, 
                self.onError,
                lambda stanza: self.onTimeout(stanza,node,msg,recpt))
        self.stream.send(ps)
       
    def onSuccess(self,stanza):
        logger.info("operation successful.")
        self.disconnect()
        return True

    def onError(self,stanza):
        errorNode = stanza.get_error()
        logger.error("error type = %s" % errorNode.get_type())
        logger.error("error message = %s" % errorNode.get_message())
        logger.info("disconnecting")
        self.disconnect()
        sys.exit(1)
        return True

    def onTimeout(self,stanza,node,msg,recpt):
        logger.info("operation timed out.  Trying again...")
        if self.counter < self.max_attempts:
            self.counter = self.counter + 1
            self.sendMessage(node,msg,recpt)
        else:
            print("Reached max_attempts. Disconnecting...")
            self.disconnect()
            sys.exit(1)
        return True

# Check for file
if not opts.file:
   print("You must supply a file to send")
   sys.exit(1)

# Debug the memory
libxml2.debugMemory(1)

# Instantiate the client
# append UUID to resource to guarantee that it is non-empty and unique
jid=JID(myusername+"@"+opts.server+"/"+opts.resource)
client = MyClient(jid,mypassword,opts.max_attempts,onStart)

# Connect
logger.info("connecting...")
client.connect()

try:
    client.loop(1)
except KeyboardInterrupt:
    print("disconnecting...")
    client.disconnect()

# vi: sts=4 et sw=4
